# (C) Copyright 2021 by Darren Rowland (advancedthermo.com). All rights reserved.
using Test
using AdvancedThermo
using CSV
using DiffResults

@testset "αʳ(helium)" begin
    helium = AdvancedThermo.get_fluid_component("helium")
    reltol = 1.e-14
    # benchmark values from TREND 4.0 (GERG-2008)
    αʳ = AdvancedThermo.αʳ(helium,50.0,300.0)
    @test abs(αʳ-5.97611193523994E-04)/αʳ<reltol
    αʳ = AdvancedThermo.αʳ(helium,20000.0,100.0)
    @test abs(αʳ-2.709103301953120E-01)/αʳ<reltol
end
@testset "αʳ(methane)" begin
    methane = AdvancedThermo.get_fluid_component("methane")
    reltol = 1.e-14
    # benchmark values from TREND 4.0 (GERG-2008)
    αʳ = AdvancedThermo.αʳ(methane,50.0,300.0)
    @test abs(αʳ+2.10257286702117E-03)/αʳ<reltol
    αʳ = AdvancedThermo.αʳ(methane,20000.0,100.0)
    @test abs(αʳ+5.12352321835192)/αʳ<reltol
end
@testset "αʳ(nitrogen)" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    reltol = 1.e-14
    # benchmark values from TREND 4.0 (GERG-2008)
    αʳ = AdvancedThermo.αʳ(nitrogen,50.0,300.0)
    @test abs(αʳ+2.22523038577245E-04)/αʳ<reltol
    αʳ = AdvancedThermo.αʳ(nitrogen,20000.0,100.0)
    @test abs(αʳ+2.21183558859261)/αʳ<reltol
end
@testset "αʳ(methane+helium)" begin
    methane = AdvancedThermo.get_fluid_component("methane")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([methane,helium])
    reltol = 1.e-14
    x = [0.7, 0.3]
    # benchmark values from TREND 4.0 (GERG-2008)
    αʳ = AdvancedThermo.αʳ(mixture,1000.0,300.0,x)
    @test abs(αʳ+1.49341854944043E-02)/αʳ<reltol
    αʳ = AdvancedThermo.αʳ(mixture,20000.0,400.0,x)
    @test abs(αʳ-2.36418364603531E-01)/αʳ<1.e-11
    αʳ = AdvancedThermo.αʳ(mixture,20000.0,200.0,x)
    @test abs(αʳ+5.20295189324906E-01)/αʳ<reltol
    αʳ = AdvancedThermo.αʳ(mixture,20000.0,700.0,x)
    @test abs(αʳ-5.10461374891548E-01)/αʳ<1.e-10
end
@testset "αʳ(nitrogen+helium)" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([nitrogen,helium])
    x = [0.7, 0.3]
    # benchmark values from TREND 4.0 (GERG-2008)
    αʳ = AdvancedThermo.αʳ(mixture,1000.0,300.0,x)
    @test abs(αʳ-7.57385600385365E-03)/αʳ<1.e-8
    αʳ = AdvancedThermo.αʳ(mixture,20000.0,400.0,x)
    @test abs(αʳ-4.90263530771741E-01)/αʳ<1.e-12
end
@testset "entropy (nitrogen+helium)" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([nitrogen,helium])
    x = [0.7, 0.3]
    # benchmark values from TREND 4.0 (GERG-2008)
    s = AdvancedThermo.s(mixture,1000.0,300.0,x)
    @test abs(s-(-21.682667884))<1.e-3
end
@testset "pressure (nitrogen+helium)" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([nitrogen,helium])
    x = [0.7, 0.3]
    # benchmark values from TREND 4.0 (GERG-2008)
    p = AdvancedThermo.p(mixture,1000.0,300.0,x)
    @test abs(p-2.514507019437e6)<5.e-3
end
@testset "derivatives of a" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([nitrogen,helium])
    x = [0.7, 0.3]
    result = AdvancedThermo.derivatives_of_a(mixture,1000.0,300.0,x)
    # Benchmark values from TREND 4.0
    # helmholtz energy a
    @test abs(DiffResults.value(result)-3991.1843584)<0.1
    # p = -∂a/∂v
    @test abs(-DiffResults.gradient(result)[1]-2.514507019437e6)<5.e-3
    # s = -∂a/∂T
    @test abs(-DiffResults.gradient(result)[2]-(-21.682667884))<5.e-3
    # ∂p/∂T_v = -∂²a/∂v∂T
    @test abs(-DiffResults.hessian(result)[1,2]-0.00860851125970e6)<1.e-4
    # volume expansivity = 1/v * ∂v/∂T_p = 1/v * -(∂p/∂T)_v/(∂p/∂v)_T
    v_exp = -1000.0*DiffResults.hessian(result)[1,2]/DiffResults.hessian(result)[1,1]
    @test abs(v_exp-0.003392906770418)<1.e-8 
end
@testset "derivatives of nαʳ" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([nitrogen,helium])
    x = [0.7, 0.3]
    ρ = 1000.0
    T = 300.0
    lnϕ = AdvancedThermo.lnϕ(mixture,ρ,T,x)
    @test abs(exp(lnϕ[1])-0.998810946)<1.e-9
    @test abs(exp(lnϕ[2])-1.028530046)<1.e-9
    methane = AdvancedThermo.get_fluid_component("methane")
    m = AdvancedThermo.create_mixture([methane,helium])
    V = inv(ρ)
    n = copy(x)
    result = AdvancedThermo.derivatives_of_nαʳ(m,V,T,n)
    # (∂²[nαʳ]/∂n₁²)
    @test abs(DiffResults.hessian(result)[3,3]-(-7.5391947562983813e-2))<1.e-12
    # (∂²[nαʳ]/∂n₂²)
    @test abs(DiffResults.hessian(result)[4,4]-(6.8649146967865726e-3))<1.e-12
    # (∂²[nαʳ]/∂n₁∂n₂)
    @test abs(DiffResults.hessian(result)[3,4]-(2.2321878662946937e-2))<1.e-12
    #Z = AdvancedThermo.Z(m,ρ,T,x)
    Zresult = AdvancedThermo.derivatives_of_Z(m,V,T,n)
    @test abs(DiffResults.value(Zresult)-(1-0.01419994899771645))<1.e-12
    Presult = AdvancedThermo.derivatives_of_p(m,V,T,n)
    @test abs(DiffResults.value(Presult)-2.4589220764971175e6)<1.e-9
    @test abs(DiffResults.gradient(Presult)[1]-(-2.4271215309896464e9))<1.e-6
    @test abs(DiffResults.gradient(Presult)[2]-(8.6559404688418290e3))<1.e-12
    @test abs(DiffResults.gradient(Presult)[3]-(2.3794078273837836e6))<1.e-9
    @test abs(DiffResults.gradient(Presult)[4]-(2.5384535060699927e6))<1.e-9

    ∂p∂V = DiffResults.gradient(Presult)[1]
    ∂p∂n₁ = DiffResults.gradient(Presult)[3]
    ∂p∂n₂ = DiffResults.gradient(Presult)[4]
    R = 8.314472
    ∂p∂n = DiffResults.gradient(Presult)[3:end]
    # ∂lnϕ∂n = DiffResults.hessian(result)[3:end,3:end] .+ 1 .+
    #               ∂p∂n*∂p∂n'./(R*T*∂p∂V)
    ∂lnϕ∂n = AdvancedThermo.∂lnϕ∂n(m,V,T,n)
    @test abs(∂lnϕ∂n[1,1]-(-1.0561415238920291e-2))<1.e-14
    @test abs(∂lnϕ∂n[1,2]-(2.4643302224147678e-2))<1.e-14
    @test abs(∂lnϕ∂n[2,2]-(-5.7501038523011472e-2))<1.e-14
end
@testset "gibbs energy (nitrogen+helium)" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([nitrogen,helium])
    x = [0.7, 0.3]
    # benchmark values from TREND 4.0 (GERG-2008)
    g = AdvancedThermo.g(mixture,1000.0,300.0,x)
    @test abs(g-6505.6913733708)<1.e-1
end
@testset "density solver (nitrogen)" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    mixture = AdvancedThermo.create_mixture([nitrogen])
    # benchmark values from TREND 4.0 (GERG-2008)
    @test abs(AdvancedThermo.ρ(mixture,1.e4,100.0,[1.0])-12.0505465245949)<1.e-7
    @test abs(AdvancedThermo.ρ(mixture,1.e6,100.0,[1.0])-24661.1684674037)<1.e-7
    @test abs(AdvancedThermo.ρ(mixture,1.e6,115.0,[1.0])-1223.49168349671)<1.e-7
    @test abs(AdvancedThermo.ρ(mixture,1.e7,115.0,[1.0])-23571.4077503479)<1.e-7
    @test abs(AdvancedThermo.ρ(mixture,2.e6,126.0,[1.0])-2500.28432820219)<1.e-7
    @test abs(AdvancedThermo.ρ(mixture,3.39e6,126.0,[1.0])-14332.0445222086)<2.e-6
    @test abs(AdvancedThermo.ρ(mixture,1.e7,126.0,[1.0])-21321.8153441504)<1.e-7

end
@testset "density solver (nitrogen+helium)" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([nitrogen,helium])
    x = [0.7, 0.3]
    # benchmark values from TREND 4.0 (GERG-2008)
    @test abs(AdvancedThermo.ρ(mixture,1.e4,300.0,x)-4.0089604347598)<1.e-6
    @test abs(AdvancedThermo.ρ(mixture,1.e4,100.0,x)-12.0356259880732)<1.e-8
    @test abs(AdvancedThermo.ρ(mixture,1.e9,300.0,x)-46376.8316627108)<1.e-8
end
@testset "Rachford-Rice solver" begin
    # adapted from 'Advanced Thermodynamics - Lecture 16 (2019)'
    z = [0.0191,0.0433,0.7046,0.1851,0.0479]
    K = [0.434,8.53,2.9696,0.3038,0.0593]
    (x,y,β) = AdvancedThermo.solve_xyβ(z,K)
    x_ref = [0.0368, 0.0058, 0.2633, 0.4541, 0.2399]
    y_ref = [0.0160, 0.0499, 0.7820, 0.1379, 0.0142]
    β_ref = 0.85078
    tol = 1.e-4
    for i = 1:length(x_ref)
        @test abs(x[i]-x_ref[i])<tol
        @test abs(y[i]-y_ref[i])<tol
    end
    @test abs(β-β_ref)<tol
end
@testset "pT flash (methane+helium)" begin
    methane = AdvancedThermo.get_fluid_component("methane")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([methane,helium])
    (x,y,β) = AdvancedThermo.flash_helium_binary(mixture,2.e6,90.0,[0.5,0.5])
    # benchmark values from TREND 4.0
    x_ref = [0.999896, 0.000104]
    y_ref = [0.007754, 0.992246]
    β_ref = 0.503856
    tol = 1.e-6
    for i = 1:length(x_ref)
        @test abs(x[i]-x_ref[i])<tol
        @test abs(y[i]-y_ref[i])<tol
    end
    @test abs(β-β_ref)<tol
end
@testset "pT flash (methane+helium)" begin
    methane = AdvancedThermo.get_fluid_component("methane")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([methane,helium])
    (x,y,β) = AdvancedThermo.flash_helium_binary(mixture,3.e6,90.0,[0.5,0.5])
    # benchmark values from TREND 4.0
    x_ref = [0.999847, 0.000153]
    y_ref = [0.006025, 0.993975]
    β_ref = 0.502955
    tol = 1.e-6
    for i = 1:length(x_ref)
        @test abs(x[i]-x_ref[i])<tol
        @test abs(y[i]-y_ref[i])<tol
    end
    @test abs(β-β_ref)<tol
end
@testset "pT flash (nitrogen+helium)" begin
    nitrogen = AdvancedThermo.get_fluid_component("nitrogen")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([nitrogen,helium])
    # benchmark values from TREND 4.0
    x_ref = [0.998260, 0.001740]
    y_ref = [0.139618, 0.860382]
    β_ref = 0.580288
    (x,y,β) = AdvancedThermo.flash_initial_K(mixture,3.e6,90.0,[0.5,0.5],y_ref./x_ref)
    tol = 1.e-6
    for i = 1:length(x_ref)
        @test abs(x[i]-x_ref[i])<tol
        @test abs(y[i]-y_ref[i])<tol
    end
    @test abs(β-β_ref)<tol
end
# @testset "pTxy data (methane+helium)" begin
#     methane = AdvancedThermo.get_fluid_component("methane")
#     helium = AdvancedThermo.get_fluid_component("helium")
#     mixture = AdvancedThermo.create_mixture([methane,helium])
#     file = "pTxy_methane_helium.csv"
#     for row in CSV.File(file)
#         z = [0.5*(row.x1+row.y1), 0.5*(row.x2+row.y2)]
#         K = [row.y1/row.x1, row.y2/row.x2]
#         try 
#             (x,y,β) = AdvancedThermo.flash_initial_K(mixture,row.p*1e3,row.T,z,K)
#             if β==0.0 || β==1.0
#                 @info "Single phase - $(row.Reference) datum $(row.n)"
#             end
#         catch e
#             @info "Internal error - $(row.Reference) datum $(row.n)"
#             println(e.msg)
#         end
#     end
# end
@testset "critical parameters (helium)" begin
    T = 300.0
    p = 1.e5
    x = [1.0]
    helium = AdvancedThermo.get_fluid_component("helium")
    m = AdvancedThermo.create_mixture([helium])
    da = AdvancedThermo.derivatives_of_a(m,AdvancedThermo.ρ(m,p,T,x),T,x)
    
    pc = 0.93e6
    Tc = 13.0
    ρc = 28.4e3
    helium_new = AdvancedThermo.get_fluid_component("helium")
    helium_new = AdvancedThermo.new_critical!(helium_new,pc,Tc,ρc)
    m_new = AdvancedThermo.create_mixture([helium_new])
    da_new = AdvancedThermo.derivatives_of_a(m_new,AdvancedThermo.ρ(m_new,p,T,x),T,x)
    
    # compare Helmholtz energy and derivatives (new vs. old)
    tol = 1.e-13
    val = DiffResults.value(da)
    val_new = DiffResults.value(da_new)
    grad = DiffResults.gradient(da)
    grad_new = DiffResults.gradient(da_new)
    hess = DiffResults.hessian(da)
    hess_new = DiffResults.hessian(da_new)
    @test abs((val-val_new)/val)<tol
    @test abs((grad[1]-grad_new[1])/grad[1])<tol
    @test abs((grad[2]-grad_new[2])/grad[2])<tol
    @test abs((grad[3]-grad_new[3])/grad[3])<tol
    @test abs((hess[1,1]-hess_new[1,1])/hess[1,1])<tol
    @test abs((hess[1,2]-hess_new[1,2])/hess[1,2])<tol
    @test abs((hess[1,3]-hess_new[1,3])/hess[1,3])<tol
    @test abs((hess[2,2]-hess_new[2,2])/hess[2,2])<tol
    @test abs((hess[2,3]-hess_new[2,3])/hess[2,3])<tol
    @test abs((hess[3,3]-hess_new[3,3])/hess[3,3])<tol
end
@testset "critical parameters (methane)" begin
    T = 300.0
    p = 1.e5
    x = [1.0]
    methane = AdvancedThermo.get_fluid_component("methane")
    m = AdvancedThermo.create_mixture([methane])
    da = AdvancedThermo.derivatives_of_a(m,AdvancedThermo.ρ(m,p,T,x),T,x)
    
    pc = 5.0e6
    Tc = 192.0
    ρc = 10.0e3
    methane_new = AdvancedThermo.get_fluid_component("methane")
    methane_new = AdvancedThermo.new_critical!(methane_new,pc,Tc,ρc)
    m_new = AdvancedThermo.create_mixture([methane_new])
    da_new = AdvancedThermo.derivatives_of_a(m_new,AdvancedThermo.ρ(m_new,p,T,x),T,x)
    
    # compare Helmholtz energy and derivatives (new vs. old)
    tol = 5.e-13
    val = DiffResults.value(da)
    val_new = DiffResults.value(da_new)
    grad = DiffResults.gradient(da)
    grad_new = DiffResults.gradient(da_new)
    hess = DiffResults.hessian(da)
    hess_new = DiffResults.hessian(da_new)
    @test abs((val-val_new)/val)<tol
    @test abs((grad[1]-grad_new[1])/grad[1])<tol
    @test abs((grad[2]-grad_new[2])/grad[2])<tol
    @test abs((grad[3]-grad_new[3])/grad[3])<tol
    @test abs((hess[1,1]-hess_new[1,1])/hess[1,1])<tol
    @test abs((hess[1,2]-hess_new[1,2])/hess[1,2])<tol
    @test abs((hess[1,3]-hess_new[1,3])/hess[1,3])<tol
    @test abs((hess[2,2]-hess_new[2,2])/hess[2,2])<tol
    @test abs((hess[2,3]-hess_new[2,3])/hess[2,3])<tol
    @test abs((hess[3,3]-hess_new[3,3])/hess[3,3])<tol
end
function calculate_statistics(dev)
    n = length(dev)
    AAD = sum(abs.(dev))/n
    Bias = sum(dev)/n
    RMS = sqrt(sum(dev.^2)/n)
    Max = maximum(dev)>-minimum(dev) ? maximum(dev) : minimum(dev)
    return (AAD, Bias, RMS, Max)
end
# @testset "Hernandez-Gomez density (methane+helium)" begin
#     methane = AdvancedThermo.get_fluid_component("methane")
#     helium = AdvancedThermo.get_fluid_component("helium")
#     mixture = AdvancedThermo.create_mixture([methane,helium])
#     file = "pdT_methane_helium.csv"
#     dev = zeros(283)
#     i = 0
#     for row in CSV.File(file)
#         p = row.p*1.e6
#         T = row.T
#         x = [row.x1, 1-row.x1]
#         ρ = AdvancedThermo.ρ(mixture,p,T,x)
#         #@info ρ, row.d, 100*(row.d-ρ)/ρ
#         i += 1
#         dev[i] = 100*(row.d-ρ)/ρ
#     end
#     # Print summary statistics
#     (AAD95,Bias95,RMS95,Max95) = calculate_statistics(dev[1:60])
#     (AAD90,Bias90,RMS90,Max90) = calculate_statistics(dev[61:119])
#     (AAD50,Bias50,RMS50,Max50) = calculate_statistics(dev[120:283])

#     println("Statistic      0.95 CH4/0.05 He               0.9 CH4/0.1 He               0.5 CH4/0.5 He")
#     println("AAD            $AAD95               $AAD90             $AAD50")
#     println("Bias           $Bias95              $Bias90            $Bias50")
#     println("RMS            $RMS95               $RMS90             $RMS50")
#     println("Max            $Max95               $Max90             $Max50")
# end
# @testset "Hernandez-Gomez density - critical constants (methane+helium)" begin
#     methane = AdvancedThermo.get_fluid_component("methane")
#     pc = 0.93e6
#     Tc = 13.0
#     ρc = 28.4e3
#     helium = AdvancedThermo.get_fluid_component("helium")
#     helium = AdvancedThermo.new_critical!(helium,pc,Tc,ρc)
#     mixture = AdvancedThermo.create_mixture([methane,helium])
#     mixture.γv[1,2] = mixture.γv[2,1] = 1.0
#     mixture.γT[1,2] = mixture.γT[2,1] = 1.0

#     file = "pdT_methane_helium.csv"
#     dev = zeros(283)
#     i = 0
#     for row in CSV.File(file)
#         p = row.p*1.e6
#         T = row.T
#         x = [row.x1, 1-row.x1]
#         ρ = AdvancedThermo.ρ(mixture,p,T,x)
#         i += 1
#         dev[i] = 100*(row.d-ρ)/ρ
#     end
#     # Print summary statistics
#     AAD95 = sum(abs.(dev[1:60]))/60
#     Bias95 = sum(dev[1:60])/60
#     RMS95 = sqrt(sum(dev[1:60].^2)/60)
#     Max95 = maximum(dev[1:60])>-minimum(dev[1:60]) ? maximum(dev[1:60]) : minimum(dev[1:60])
#     AAD90 = sum(abs.(dev[61:119]))/59
#     Bias90 = sum(dev[61:119])/59
#     RMS90 = sqrt(sum(dev[61:119].^2)/59)
#     Max90 = maximum(dev[61:119])>-minimum(dev[61:119]) ? maximum(dev[61:119]) : minimum(dev[61:119])
#     AAD50 = sum(abs.(dev[120:283]))/164
#     Bias50 = sum(dev[120:283])/164
#     RMS50 = sqrt(sum(dev[120:283].^2)/164)
#     Max50 = maximum(dev[120:283])>-minimum(dev[120:283]) ? maximum(dev[120:283]) : minimum(dev[120:283])

#     println("Statistic      0.95 CH4/0.05 He               0.9 CH4/0.1 He               0.5 CH4/0.5 He")
#     println("AAD            $AAD95               $AAD90             $AAD50")
#     println("Bias           $Bias95              $Bias90            $Bias50")
#     println("RMS            $RMS95               $RMS90             $RMS50")
#     println("Max            $Max95               $Max90             $Max50")
# end
# @testset "pTxy data (methane+helium(modified Tc/ρc))" begin
#     methane = AdvancedThermo.get_fluid_component("methane")
#     pc = 0.93e6
#     Tc = 13.0
#     ρc = 28.4e3
#     helium = AdvancedThermo.get_fluid_component("helium")
#     helium = AdvancedThermo.new_critical!(helium,pc,Tc,ρc)
#     mixture = AdvancedThermo.create_mixture([methane,helium])
#     mixture.γv[1,2] = mixture.γv[2,1] = 1.0
#     mixture.γT[1,2] = mixture.γT[2,1] = 1.0
#     file = "pTxy_methane_helium.csv"
#     for row in CSV.File(file)
#         z = [0.5*(row.x1+row.y1), 0.5*(row.x2+row.y2)]
#         K = [row.y1/row.x1, row.y2/row.x2]
#         try 
#             (x,y,β) = AdvancedThermo.flash_initial_K(mixture,row.p*1e3,row.T,z,K)
#             if β==0.0 || β==1.0
#                 @info "Single phase - $(row.Reference) datum $(row.n)"
#             end
#         catch e
#             @info "Internal error - $(row.Reference) datum $(row.n)"
#             println(e.msg)
#         end
#     end
# end
@testset "pruned pTxy data (methane+helium)" begin
    methane = AdvancedThermo.get_fluid_component("methane")
    helium = AdvancedThermo.get_fluid_component("helium")
    mixture = AdvancedThermo.create_mixture([methane,helium])
    file = "pTxy_methane_helium_pruned.csv"
    devK1 = zeros(109)
    devK2 = zeros(109)
    i = 0
    for row in CSV.File(file)
        z = [0.5*(row.x1+row.y1), 0.5*(row.x2+row.y2)]
        K = [row.y1/row.x1, row.y2/row.x2]
        try 
            (x,y,β) = AdvancedThermo.flash_initial_K(mixture,row.p*1e3,row.T,z,K)
            if β==0.0 || β==1.0
                # don't expect to encounter this with the pruned dataset
                @info "Single phase - $(row.Reference) datum $(row.n)"
            else
                i += 1
                devK1[i] = log(K[1]) - log(y[1]/x[1])
                devK2[i] = log(K[2]) - log(y[2]/x[2])
                if abs(devK2[i])>10
                    @info i
                end
                #@info "K1=$(K[1]) K2=$(K[2]); y1=$(y[1]) x1=$(x[1]) K1=$(y[1]/x[1]) K2=$(y[2]/x[2])"
            end
        catch e
            # don't expect to encounter this with the pruned dataset
            @info "Internal error - $(row.Reference) datum $(row.n)"
            println(e.msg)
        end
    end
    (AADK1,BiasK1,RMSK1,MaxK1) = calculate_statistics(devK1)
    (AADK2,BiasK2,RMSK2,MaxK2) = calculate_statistics(devK2)
    println("Statistic      lnK(methane)               lnK(helium)")
    println("AAD            $AADK1               $AADK2")
    println("Bias           $BiasK1              $BiasK2")
    println("RMS            $RMSK1               $RMSK2")
    println("Max            $MaxK1               $MaxK2")
end
@testset "pruned pTxy data (methane+helium(modified Tc/ρc))" begin
    methane = AdvancedThermo.get_fluid_component("methane")
    pc = 0.93e6
    Tc = 13.0
    ρc = 28.4e3
    helium = AdvancedThermo.get_fluid_component("helium")
    helium = AdvancedThermo.new_critical!(helium,pc,Tc,ρc)
    mixture = AdvancedThermo.create_mixture([methane,helium])
    mixture.γv[1,2] = mixture.γv[2,1] = 1.0
    mixture.γT[1,2] = mixture.γT[2,1] = 1.0
    file = "pTxy_methane_helium_pruned.csv"
    devK1 = zeros(109)
    devK2 = zeros(109)
    i = 0
    for row in CSV.File(file)
        z = [0.5*(row.x1+row.y1), 0.5*(row.x2+row.y2)]
        K = [row.y1/row.x1, row.y2/row.x2]
        try 
            (x,y,β) = AdvancedThermo.flash_initial_K(mixture,row.p*1e3,row.T,z,K)
            if β==0.0 || β==1.0
                # don't expect to encounter this with the pruned dataset
                @info "Single phase - $(row.Reference) datum $(row.n)"
            else
                i += 1
                devK1[i] = log(K[1]) - log(y[1]/x[1])
                devK2[i] = log(K[2]) - log(y[2]/x[2])
                if abs(devK2[i])>10
                    @info i
                end
                #@info "K1=$(K[1]) K2=$(K[2]); y1=$(y[1]) x1=$(x[1]) K1=$(y[1]/x[1]) K2=$(y[2]/x[2])"
            end
        catch e
            # don't expect to encounter this with the pruned dataset
            @info "Internal error - $(row.Reference) datum $(row.n)"
            println(e.msg)
        end
    end
    (AADK1,BiasK1,RMSK1,MaxK1) = calculate_statistics(devK1)
    (AADK2,BiasK2,RMSK2,MaxK2) = calculate_statistics(devK2)
    println("Statistic      lnK(methane)               lnK(helium)")
    println("AAD            $AADK1               $AADK2")
    println("Bias           $BiasK1              $BiasK2")
    println("RMS            $RMSK1               $RMSK2")
    println("Max            $MaxK1               $MaxK2")
end
# @testset "Hernandez-Gomez density - critical constants, arithmetic γ (methane+helium)" begin
#     methane = AdvancedThermo.get_fluid_component("methane")
#     pc = 0.93e6
#     Tc = 13.0
#     ρc = 28.4e3
#     helium = AdvancedThermo.get_fluid_component("helium")
#     helium = AdvancedThermo.new_critical!(helium,pc,Tc,ρc)
#     mixture = AdvancedThermo.create_mixture([methane,helium])
#     (γv12, γT12) = AdvancedThermo.arithmetic_γ(methane, helium)
#     mixture.γv[1,2] = mixture.γv[2,1] = γv12
#     mixture.γT[1,2] = mixture.γT[2,1] = γT12

#     file = "pdT_methane_helium.csv"
#     dev = zeros(283)
#     i = 0
#     for row in CSV.File(file)
#         p = row.p*1.e6
#         T = row.T
#         x = [row.x1, 1-row.x1]
#         ρ = AdvancedThermo.ρ(mixture,p,T,x)
#         i += 1
#         dev[i] = 100*(row.d-ρ)/ρ
#     end
#     # Print summary statistics
#     AAD95 = sum(abs.(dev[1:60]))/60
#     Bias95 = sum(dev[1:60])/60
#     RMS95 = sqrt(sum(dev[1:60].^2)/60)
#     Max95 = maximum(dev[1:60])>-minimum(dev[1:60]) ? maximum(dev[1:60]) : minimum(dev[1:60])
#     AAD90 = sum(abs.(dev[61:119]))/59
#     Bias90 = sum(dev[61:119])/59
#     RMS90 = sqrt(sum(dev[61:119].^2)/59)
#     Max90 = maximum(dev[61:119])>-minimum(dev[61:119]) ? maximum(dev[61:119]) : minimum(dev[61:119])
#     AAD50 = sum(abs.(dev[120:283]))/164
#     Bias50 = sum(dev[120:283])/164
#     RMS50 = sqrt(sum(dev[120:283].^2)/164)
#     Max50 = maximum(dev[120:283])>-minimum(dev[120:283]) ? maximum(dev[120:283]) : minimum(dev[120:283])

#     println("Statistic      0.95 CH4/0.05 He               0.9 CH4/0.1 He               0.5 CH4/0.5 He")
#     println("AAD            $AAD95               $AAD90             $AAD50")
#     println("Bias           $Bias95              $Bias90            $Bias50")
#     println("RMS            $RMS95               $RMS90             $RMS50")
#     println("Max            $Max95               $Max90             $Max50")
# end

@testset "pruned pTxy data (methane+helium(modified Tc/ρc), arithmetic γ)" begin
    methane = AdvancedThermo.get_fluid_component("methane")
    pc = 0.93e6
    Tc = 13.0
    ρc = 28.4e3
    helium = AdvancedThermo.get_fluid_component("helium")
    helium = AdvancedThermo.new_critical!(helium,pc,Tc,ρc)
    mixture = AdvancedThermo.create_mixture([methane,helium])
    (γv12, γT12) = AdvancedThermo.arithmetic_γ(methane, helium)
    mixture.γv[1,2] = mixture.γv[2,1] = γv12
    mixture.γT[1,2] = mixture.γT[2,1] = γT12
    file = "pTxy_methane_helium_pruned.csv"
    devK1 = zeros(109)
    devK2 = zeros(109)
    i = 0
    for row in CSV.File(file)
        z = [0.5*(row.x1+row.y1), 0.5*(row.x2+row.y2)]
        K = [row.y1/row.x1, row.y2/row.x2]
        try 
            (x,y,β) = AdvancedThermo.flash_initial_K(mixture,row.p*1e3,row.T,z,K)
            if β==0.0 || β==1.0
                # don't expect to encounter this with the pruned dataset
                @info "Single phase - $(row.Reference) datum $(row.n)"
            else
                i += 1
                devK1[i] = log(K[1]) - log(y[1]/x[1])
                devK2[i] = log(K[2]) - log(y[2]/x[2])
                if abs(devK2[i])>10
                    @info i
                end
                #@info "K1=$(K[1]) K2=$(K[2]); y1=$(y[1]) x1=$(x[1]) K1=$(y[1]/x[1]) K2=$(y[2]/x[2])"
            end
        catch e
            # don't expect to encounter this with the pruned dataset
            @info "Internal error - $(row.Reference) datum $(row.n)"
            println(e.msg)
        end
    end
    (AADK1,BiasK1,RMSK1,MaxK1) = calculate_statistics(devK1)
    (AADK2,BiasK2,RMSK2,MaxK2) = calculate_statistics(devK2)
    println("Statistic      lnK(methane)               lnK(helium)")
    println("AAD            $AADK1               $AADK2")
    println("Bias           $BiasK1              $BiasK2")
    println("RMS            $RMSK1               $RMSK2")
    println("Max            $MaxK1               $MaxK2")
end
# @testset "pTxy data (methane+helium)" begin
#     methane = AdvancedThermo.get_fluid_component("methane")
#     helium = AdvancedThermo.get_fluid_component("helium")
#     mixture = AdvancedThermo.create_mixture([methane,helium])
#     file = "flash_benchmark_GERG2008_methane+helium.csv"
#     correct = 0
#     errors = 0
#     for row in CSV.File(file)
#         z = [row.z1, 1-row.z1]
#         try 
#             #(x,y,β) = AdvancedThermo.flash(mixture,row.p*1e3,row.T,z)
#             #(x,y,β) = AdvancedThermo.flash_helium_binary(mixture,row.p*1e3,row.T,z)
#             (x,y,β) = AdvancedThermo.flash_initial_K(mixture,row.p*1e3,row.T,z,[row.y1/row.x1, (1-row.y1)/(1-row.x1)])
#             if β==0.0 || β==1.0
#                 @info "Single phase - row $(row.n)"
#             elseif abs(β-row.beta)>2.e-6
#                 @info "Incorrect β - row $(row.n)"
#             elseif abs(x[1]-row.x1)>2.e-6
#                 @info "Incorrect x1 - row $(row.n)"
#             elseif abs(y[1]-row.y1)>2.e-6
#                 @info "Incorrect y1 - row $(row.n)"
#             else
#                 correct += 1
#             end
#         catch e
#             @info "Internal error - row $(row.n)"
#             println(e.msg)
#             errors += 1
#         end
#     end
#     @info "Correct: $correct,  Incorrect: $(56-correct-errors),  Errors: $errors"
# end
@testset "stability (methane+helium)" begin
    methane = AdvancedThermo.get_fluid_component("methane")
    helium = AdvancedThermo.get_fluid_component("helium")
    m = AdvancedThermo.create_mixture([methane,helium])
    T = 188.01
    p = 4819.4e3
    z1 = 0.9757
    z = [z1, 1-z1]
    (W, tpd, tm) = AdvancedThermo.stability_helium_binary(m,p,T,z)
    w = W./sum(W)
    @test abs(w[1]-0.989336)<1.e-6
    @test abs(tm-(-0.0031216655))<1.e-8
end
# @testset "stability benchmark (methane+helium)" begin
#     methane = AdvancedThermo.get_fluid_component("methane")
#     helium = AdvancedThermo.get_fluid_component("helium")
#     mixture = AdvancedThermo.create_mixture([methane,helium])
#     file = "flash_benchmark_GERG2008_methane+helium.csv"
#     for row in CSV.File(file)
#         z = [row.z1, 1-row.z1]
#         p = row.p*1.e3
#         T = row.T
#         try
#             (W,tpd,tm) = AdvancedThermo.stability_helium_binary(mixture,p,T,z)
#             if tm >= 0.0
#                 @info "tm>0 - row $(row.n)"
#             end
#         catch e
#             @info "Internal error - row $(row.n)"
#         end
#     end
# end